//////////////////////////////////////////////////////////////////////
// Math64.cpp: implementation of the 64-bit integer math extender.
//
// This example demonstrates using the CWilExtenderT template and 
// CExtender base class to create an a C++ based WIL extender. The 
// extender template handles most of the WIL boilerplate 
// requirements so the developer can focus on the details unique to 
// the specific extender.  The extender template does require the
// a few basic steps:
//
// 1.  Define your extender class interface as derived from the 
//     CWilExtenderT template class.  For example, to declare the
//     "CSpiffy" extender class you would do follows: 
//     class CSpiffy: public CWilExtenderT<CSpiffy> {...}
//     (see Math64.h)
//
// 2.  Add implementations of the required "Identity", 
//     "InterfaceVersion", "Initalization", "LoadTable", and
//     "ShutDown" member functions to your class.  The simplest
//     way to do this is to copy the member definitions at the 
//     bottom of this source file into your class implementation
//     changing the class names in the appropriate places.  Don't
//     forget to add the member declarations for those members to 
//     your class's class declaration (usually in a header file), 
//     as well.
//                                                                                                                                                                                                                      
// 3.  Make sure you create an instance of your class at the top
//     of your main source file. For example, to create an instance 
//     of the "CSpiffy" class use:
//
//     CSpiffy Spiffy;
//
// 4.  Create a command table for each WIL function and constant
//     implemented by you extender. The template expects a static
//     array of type EntryT elements.  Each array element represents
//     one WIL function or WIL constant implemented by your extender.
//
//     Because the table is static it must be defined outside class
//     members.  For example, the definition the table for the CSpiffy 
//     extender would look like the following:
//
//     CWilExtenderT<CSpiffy>::EntryT<CSpiffy> CWilExtenderT<CSpiffy>::m_ExtTable[] = 
//     {...};
//
//     The bit between the braces contains the descriptions of the WIL 
//     functions and constants implemented by your extender.  Each 
//     comma separated entry has the following format for functions:
//
//                       Member Ptr       Function Name  Name Chars  Required Args  Optional Args     Arg Types...
//     EntryT<CSpiffy>(&CSpiffy::Handy,   _T("sphandy"),     7,           2,            1,         PTStr,PTStr,PTInt),
//
//     The above command table entry identifies a WIL function named 
//     "spHandy" that  takes 2 required and 1 optional parameter.  The 
//     first 2 parameters are of type string and the third parameter is 
//     of type integer. The member pointer is the class member that 
//     implements the functionality exposed by the "spHandy" WIL 
//     functions.
//
//     A constant table entry looks something like this:
//
//                         Constant Name   Name Chars   Value
//    EntryT<CSpiffy>(  _T("@spWhatever"),    11,        -1 ),
//
//    The above entry implements a WIL the constant "@spWhatever"
//    with a value of -1.
//
//    View the command map of below for a detailed example.
//
// 5.  Implement class members to support each WIL extender
//     function. 
//     Note: All members referenced in the command map must 
//           accept the same five parameters and return LONG.
// 
// While C++ templates can appear difficult to decipher to those
// unfamiliar with generic programming, they offer several advantages.
// One advantage is that using templates often significantly reduces 
// the amount of source code written.  Another is they can reduce 
// the amount processing overhead caused by C++ subclassing. They 
// can also significantly reduce the size of the linker's output 
// because only called members are compiled into the output image.
//
// NOTE:
// The extender template and example are targeted for Microsoft Visual 
// C++ v9.0 (Visual Studio 2008/_MSC_VER == 1500) and newer versions 
// of the Microsoft C++ compiler. It will likely not work with 
// Mircosoft Clang, gcc, or other compilers .
//
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////
//               64-bit Integer Operator Extender - Design Goal
//
//  This extender's targeted goal is to expose 64-bit integer functions that
//  mimic the behavior of WIL 32-bit integer operators on 32-bit integers using
//  64-bit integers instead. 
//
//  The one significant difference between this extenders implementation of 
//  64-bit integer operations and WIL's 32-bit integer operations reside in the
//  the conversion of string to integers.  WIL simply allows over and under 
//  flows to occur whereas this extender generates an error if the string 
//  representation of a number causes an overflow or underflow during conversion
//  to a 64-bit integer.
//
/////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////
// Headers files as needed.
/////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"  // Standard MFST Pre-compiled header. Only include stuff that doesn't change much.
using namespace std; // Makes for less typing when using the STD/STL. 

#include "resource.h"    // Version and error string stuff.
#include "WTExtender.h"  // The extender template and base class declarations.
#include "Math64.h"      // The 64-bit math extender declared.

CMath64 Math64;   // The one and only instance of the extender.

///////////////////////////////////////////////////////////////////////////////////////
// Using the C++ array data type so create the statice array instance and initialize it.
const CWilExtenderT<CMath64>::EntryT<CMath64> CWilExtenderT<CMath64>::m_ExtTable[] =
/////////////////   Member Ptr        Function Name     Name Chars  Required Args Optional Args       Arg Types...
{EntryT<CMath64>(&CMath64::m64Add,    _T("m64add"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Sub,    _T("m64sub"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Mul,    _T("m64mul"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Div,    _T("m64div"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Mod,    _T("m64mod"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Pow,    _T("m64pow"),         6,            2,           0,           PTTSt, PTInt),
 EntryT<CMath64>(&CMath64::m64Rsh,    _T("m64rsh"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Lsh,    _T("m64lsh"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64And,    _T("m64and"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Or,     _T("m64or"),          5,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Xor,    _T("m64xor"),         6,            2,           0,           PTTSt, PTTSt),
 EntryT<CMath64>(&CMath64::m64Not,    _T("m64not"),         6,            1,           0,           PTTSt),
 EntryT<CMath64>(&CMath64::m64Conv,   _T("m64Conv"),        7,            2,           0,           PTTSt, PTInt), 
 EntryT<CMath64>(&CMath64::m64GetInfo,_T("m64GetInfo"),    10,            1,           0,           PTInt), 
 ////////////////////////////////    Constant Name      Name Chars            Value
 EntryT<CMath64>(                    _T("@m64todec"),       9,            Math64.m64ToDec),
 EntryT<CMath64>(                    _T("@m64tohex"),       9,            Math64.m64ToHex) };

//////////////////////////////////////////////////////////////////////////////////////
// Extender WIL function implementations.

/********************************************************************
 * Name: m64Add
 *
 * Purpose: Sums two 64-bit integers.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Add": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit sum as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Add(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	TCHAR szSum[65] = {0};
	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));

		// Do the math.
		INT64 nSum = nArg1 + nArg2;

		// Convert to a string.
		if ( _i64tot_s(nSum, szSum, _countof(szSum), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szSum);
}

/********************************************************************
 * Name: m64Sub
 *
 * Purpose: Subtracts two 64-bit integers.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * Notes: The second WIL function parameter is subtracted from the
 *        first WIL function parameter
 *
 * WIL Function "m64Sub": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit difference as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Sub(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError   = 0;
	TCHAR szDif[65]  = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));

		// Do the math.
		INT64 nDif = nArg1 - nArg2;

		// Convert to a string.
		if ( _i64tot_s(nDif, szDif, _countof(szDif), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szDif);
}

/********************************************************************
 * Name: m64Mul
 *
 * Purpose: Multiples two 64-bit integers.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Mul": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit product as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Mul(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError   = 0;
	TCHAR szProd[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));

		// Do the math.
		INT64 nProd = nArg1 * nArg2;

		// Convert to a string.
		if ( _i64tot_s(nProd, szProd, _countof(szProd), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szProd);
}

/********************************************************************
 * Name: m64Div
 *
 * Purpose: Divides two 64-bit integers and reports the quotient as
 *          the result.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * Notes: First argument is the dividend and the second is the 
 *        divisor.
 *
 * WIL Function "m64Div": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit quotient as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Div(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	int nExtError    = 0;
	TCHAR szQuot[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));
		if ( nArg2 == 0 )
			nExtError = IDS_DIVZERO;
		else {
			// Do the math.
			INT64 nQuot = nArg1 / nArg2;

			// Convert to a string.
			if ( _i64tot_s(nQuot, szQuot, _countof(szQuot), 10) )
				nExtError = IDS_INTERNAL;
		}
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szQuot);
}

/********************************************************************
 * Name: m64Mod
 *
 * Purpose: Divides two 64-bit integers and reports the remainser as
 *          the result.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * Notes: First argument is the dividend and the second is the 
 *        divisor.
 *
 * WIL Function "m64Mod": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit remainder as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Mod(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	TCHAR szRem[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));
		if ( nArg2 == 0 )
			nExtError = IDS_DIVZERO;
		else {
			// Do the math.
			INT64 nRem = nArg1 % nArg2;

			// Convert to a string.
			if ( _i64tot_s(nRem, szRem, _countof(szRem), 10) )
				nExtError = IDS_INTERNAL;
		}
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szRem);
}

/********************************************************************
 * Name: m64Pow
 *
 * Purpose: Raises the first argument to the power of the second.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * Notes: Can produce unacceptable results because of loss of 
 *        precision.
 *
 * WIL Function "m64Pow": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 32-bit integer.
 *
 *                        Return - 64-bit result as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Pow(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError   = 0;
	TCHAR szProd[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT nArg2 = lpvvArg[1].x;
		if ( nArg2 < 0 )  // Can't represent fractions.
			nExtError = IDS_EXPOOUTOFRANGE;
		else {
			
			// The pow std libary function can lose precision and the 
			// goal is to process 64-bit integers the same way WIL process
			// 32-bit integers. So 64-bit raising to a power is handled 
			// using straight multiplication in a Duff's Machine.
			INT64 nProd = IntPow(nArg1, nArg2);

			// Convert to a string. 
			if ( _i64tot_s(static_cast<INT64>(nProd), szProd, _countof(szProd), 10) )
				nExtError = IDS_INTERNAL;
		}
	}
	catch(CErrnoException &en ) { // Errno error.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szProd);
}

/********************************************************************
 * Name: m64Rsh
 *
 * Purpose: Bit shifts the first argument to the right the second
 *          argument number of bits.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Rsh": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit result as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Rsh(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError     = 0;
	TCHAR szShifty[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));

		// Do the shift.
		INT64 nShifty = nArg1 >> nArg2;

		// Convert to a string.
		if ( _i64tot_s(nShifty, szShifty, _countof(szShifty), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szShifty);
}

/********************************************************************
 * Name: m64Lsh
 *
 * Purpose: Bit shifts the first argument to the left the second
 *          argument number of bits.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Lsh": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit result as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Lsh(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError     = 0;
	TCHAR szShifty[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));

		// Do the shift.
		INT64 nShifty = nArg1 << nArg2;

		// Convert to a string.
		if ( _i64tot_s(nShifty, szShifty, _countof(szShifty), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szShifty);
}

/********************************************************************
 * Name: m64And
 *
 * Purpose: Performs bitwise AND operation on two 64-bit integers.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64And": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit result as a string.
 *        
 ********************************************************************/
LONG CMath64::m64And(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	TCHAR szAnd[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));

		// Do the bit twiddle.
		INT64 nAnd = nArg1 & nArg2;

		// Convert to a string.
		if ( _i64tot_s(nAnd, szAnd, _countof(szAnd), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szAnd);
}

/********************************************************************
 * Name: m64Or
 *
 * Purpose: Performs bitwise OR operation on two 64-bit integers.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Or": Arg 1 - 64-bit integer string.
 *                       Arg 2 - 64-bit integer string.
 *
 *                       Return - 64-bit result as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Or(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	TCHAR szAnd[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));

		// Do the bit twiddle.
		INT64 nAnd = nArg1 | nArg2;

		// Convert to a string.
		if ( _i64tot_s(nAnd, szAnd, _countof(szAnd), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szAnd);
}

/********************************************************************
 * Name: m64Xor
 *
 * Purpose: Performs bitwise XOR operation on two 64-bit integers.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Xor": Arg 1 - 64-bit integer string.
 *                        Arg 2 - 64-bit integer string.
 *
 *                        Return - 64-bit result as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Xor(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	TCHAR szAnd[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Convert the second  parameter.
		INT64 nArg2 = StrToInt64(ArgTStr(lpvvArg[1]));

		// Do the bit twiddle.
		INT64 nAnd = nArg1 ^ nArg2;

		// Convert to a string.
		if ( _i64tot_s(nAnd, szAnd, _countof(szAnd), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szAnd);
}

/********************************************************************
 * Name: m64Not
 *
 * Purpose: Takes the ones complement of a 64-bit integers.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Not": Arg 1 - 64-bit integer string.
 *
 *                        Return - 64-bit result as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Not(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	TCHAR szNot[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		// Convert the first parameter.
		INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]));

		// Do the bits.
		INT64 nNot = ~nArg1;

		// Convert to a string.
		if ( _i64tot_s(nNot, szNot, _countof(szNot), 10) )
			nExtError = IDS_INTERNAL;
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szNot);
}

/********************************************************************
 * Name: m64Conv
 *
 * Purpose: Converts a 64-bit integer from decimal to hexidecimal or
 *          from hexidecimal to decimal.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Conv": Arg 1 - 64-bit integer string.
 *                         Arg 2 - The conversion radix. (@m64ToDec or @m64ToHex) 
 *
 *                        Return - 64-bit decimal or hexidecimal 
 *                                 result as a string.
 *        
 ********************************************************************/
LONG CMath64::m64Conv(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nExtError  = 0;
	TCHAR szConv[65] = {0};

	try {
		CNumLocale Locale;  
	
		// Use the current user's locale for conversions.
		Locale.SetToEnvironment();
		
		INT nArg2 = lpvvArg[1].x;

		// Use second agrument to determine current radix, 
		// as well as, new radix.	
		int nFromBase = 0;
		int nToBase   = 0;
		if ( nArg2 == m64ToDec ) {
			nFromBase = m64ToHex;
			nToBase   = m64ToDec;
		}
		else if ( nArg2 == m64ToHex ) {
			nFromBase = m64ToDec;
			nToBase   = m64ToHex;
		}
		else
			nExtError = IDS_INVALIDRADIX;

		if ( nExtError == 0 ) {      
			
			// Convert the first parameter using from radix.
			INT64 nArg1 = StrToInt64(ArgTStr(lpvvArg[0]), nFromBase);

			// Convert to a string of specified radix.
			if ( _i64tot_s(nArg1, szConv, _countof(szConv), nToBase) )
				nExtError = IDS_INTERNAL;
		}			 
	}
	catch(CErrnoException &en ) { // Might be one of ours.
		if ( en.GetErrno() == ERANGE )
			nExtError = IDS_OUTOFRANGE;
		else if ( en.GetErrno() == EINVAL )
			nExtError = IDS_NOTNUMBER;
		else
			nExtError = IDS_SYSTEMERROR;
	}
	catch(exception &e) {  // STD error.
		nExtError = IDS_SYSTEMERROR;
		DllStoreErrorInfo(e.what());
	}

	// Let all other exceptions propagate.

	if ( nExtError )
		return ReturnError(RESULT_MINOR_ERROR, nExtError);
	
	return ReturnResult(szConv);
}

/********************************************************************
 * Name: m64Info
 *
 * Purpose: Reports probably not very usefull basic extender 
 *          information.
 *
 * Return: 0 on success and < 0 on failure.   
 *
 * WIL Function "m64Info": Arg 1 - number indicating requested info
 *                                 where:
 *                                 0 = version number
 *                                 1 = function count
 *                                 2 = constant count
 *
 *                        Return - version or count.  -1 if request
 *                                 not recognized.
 *        
 ********************************************************************/
LONG CMath64::m64GetInfo(HWND hWnd, HINSTANCE hInst, LPLONG lpLong, LPVIPERVAR lpvvArg, LONG nArgCount)
{
	LONG nReturn;

	INT nArg = lpvvArg[0].x;
	if ( nArg == 0 )                   // Version number.
		nReturn = (LONG)m_dwExtVersion;
	else if ( nArg == 1 ) {            // Functions.
		nReturn = 0;  
		for ( int i = 0; i < _countof(m_ExtTable); i++) 
			if ( m_ExtTable[i].m_nConstFunc == FUNCTION )
				nReturn++;
	}
	else if ( nArg == 2 ) {            // Constants.
		nReturn = 0;
		for ( int i = 0; i < _countof(m_ExtTable); i++) 
			if ( m_ExtTable[i].m_nConstFunc == CONSTANT )
				nReturn++;
	}
	else
		nReturn = -1;

	return ReturnResult(nReturn);
}

///////////////////////////////////////////////////////////////////////////////////////
// Utility memebers.

/********************************************************************
 * Name: StrToInt64 
 *
 * Purpose: Converts text to a 64-bit integer.
 *
 * Return: 64-bit integer.   
 *
 * Notes: Throws errno value when out of range error or string pointer
 *        is NULL. Throws an EINVAL errno exception when the input does 
 *        not follow the pattern [whitespace] [sign] [base digits]]. 
 *
 *        Floating point numbers are not supported.
 *
 ********************************************************************/
INT64 CMath64::StrToInt64( LPTSTR pszNumber,   // In - String to convert
								   int nBase /*=10*/)  // In - Radix to convert from.
{
	// Clear trailing WIL whitespace which includes only 
	// the space and tab charaters.
	tstring Source = pszNumber;
   string::size_type Found = Source.find_last_not_of(_T(" \t"));
	if ( Found < (Source.length()-1) && (Found != string::npos) )
		Source.erase(Found+1);
	
	// Convert with error detection. 
	// (Errno is global so used with care.)
	errno_t saved_errno = errno;
	_set_errno(0);
	LPTSTR pEnd = NULL;
	INT64 nReturn = _tcstoi64(Source.c_str(), &pEnd, nBase);
	errno_t nErrno = errno;
   _set_errno(saved_errno);
	if ( nErrno == ERANGE || nErrno == EINVAL ) 
		throw CErrnoException(nErrno); 

	// All characters must convert or its an error.
	if ( *pEnd != _T('\0') || Source.empty() ) 
		throw CErrnoException(EINVAL);   // Hijack errno's error number.

	return nReturn;
}

/********************************************************************
 * Name: IntPow
 *
 * Purpose: Raises a 64-bit integer to a positive integer power.
 *
 * Return: 64-bit integer result of base raised to exponent or -1
 *         on error.  Integer overflow returns are compiler and 
 *         processor dependent.
 *
 * Notes: Uses a Duff's Device to force loop unrolling. This is likely
 *        an unnecessary conceit given that most C++ compilers will
 *        do this type of optimization for you...
 *        
 ********************************************************************/
INT64 CMath64::IntPow( INT64 nBase,
				           int   nExponent ) 
{
	
	if ( nExponent < 1 ) {
		if ( nExponent == 0 )
			return 1;
		return -1;
	}

	register INT n = (nExponent + 7) / 8;	
	register INT64 nResult = 1;
	switch ( nExponent % 8 ) {
	 case 0: do { nResult *= nBase;
	 case 7:      nResult *= nBase;
	 case 6:      nResult *= nBase;
	 case 5:      nResult *= nBase;
	 case 4:      nResult *= nBase;
	 case 3:      nResult *= nBase;
	 case 2:      nResult *= nBase;
	 case 1:      nResult *= nBase;
				} while ( --n > 0 );
	}

	return nResult;
}

///////////////////////////////////////////////////////////////////////////////////////
// Required boilerplate extender interface members.
///////////////////////////////////////////////////////////////////////////////////////

/********************************************************************
 * Name: Identify (case 1 : WILExtenderQuery(W))
 *
 * Purpose: Case 1 is used to identify yourself to a caller.
 *          If lpvInfo is not NULL, it points to a character buffer where you may copy
 *          a string describing the extender.
 *          nCount is the size of the buffer pointed to by lpvInfo, in bytes.  When
 *          called by the WIL DLL, you can expect this to be a 256-byte buffer.
 *          The return value should be the version number of the extender.
 *
 * Return: Extender version number.
 *        
 * Notes: 
 *
 ********************************************************************/
LONG CMath64::Identity( LPVOID lpvInfo,  // information buffer
                        LONG   nCount)   // additional data
{
   // Copy the description stored as a resource.
   if ( lpvInfo && nCount > 0 ) {
		LoadMyExtString(IDS_DESCRIPTION, m_szDescription, _countof(m_szDescription));  // From resource or language file.
      lstrcpyn((LPTSTR) lpvInfo, m_szDescription, min(nCount, _countof(m_szDescription)));
   }

	// Extract the dll's base name for use as the extender's name.
	TCHAR szFullName[MAX_PATH+1];
	int nResult = ::GetModuleFileName(CExtenderBase::m_hInstance, szFullName, _countof(szFullName));
	if ( nResult && nResult < _countof(szFullName) ) 
		_tsplitpath_s(szFullName, NULL, 0, NULL, 0, m_szExtName, _countof(m_szExtName), NULL, 0);

   // Attempt to extract version info from the dll's version resource block.
   DWORD dwExtVersion = 10001; // If all else fails.
   GetProductVersionNumber(m_hInstance, dwExtVersion);
   m_dwExtVersion = dwExtVersion;
   
   return (LONG) dwExtVersion;
}

/********************************************************************
 * Name: InterfaceVersion (case 2 : WILExtenderQuery(W))
 *
 * Purpose: case 2 is used to inform the WIL DLL of your preferred interface level.
 *          lpvInfo points to an array of LONG's, indicating the interface levels
 *          supported by the WIL DLL (each element represents one level).
 *          nCount indicates the number of elements in the array pointed to by lpvInfo.
 *       
 *          The return value must be the same as one of the elements in the lpvInfo
 *          array.  Currently, the WIL DLL supports interface levels 1 thu 5, so your
 *          return value should be 1 or 2.  Newly designed extenders (such as WILX)
 *          use interface level 2 and have a WilExtender2 entry point.  However, it
 *          is possible to design an extender to support multiple interface levels.
 *         
 *          If you return an invalid value, the extender will not be loaded and the
 *          AddExtender command will fail.           
 *
 * Return: Peferred interface version number
 *        
 * Notes:
 *
 ********************************************************************/
long CMath64::InterfaceVersion( LPVOID lpvInfo,  // information buffer
                                LONG   nCount)   // additional data
{
   for (int i = 0; i < nCount; i++)
      if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL) // Found one we like.
         return INTERFACELEVEL;
      
   return -1;
}

/********************************************************************
 * Name: Initialization  (case -1 : WILExtender2(W))
 *
 * Purpose: Case -1 is called once, when your extender is loaded.  
 *          Use this member to perform your one-time initialization 
 *          tasks.  The return value should be the number of items 
 *          in your command map (functions and constants).  You may
 *          also return 0 if you are unable to determine the correct 
 *          number of items. You may return a negative value (-1) to 
 *          indicate that your extender was unable to successfully 
 *          initialize, for example, if it is running on an unsupported 
 *          platform or a dependent DLL could not be loaded.  This will
 *          cause the AddExtender command to fail.  You may additionally 
 *          provide a  string indicating the reason for the failure:
 *          lpvvArg points to a character buffer into which you may 
 *          copy an error message.
 *          "nArgCount" is the size of the buffer pointed to by lpvvArg, 
 *          in bytes. You can expect this to be a 256-byte buffer. 
 *          You can use this member implementation in your own extender 
 *          by changing the type name in used with the EntryT<> template 
 *          class referenced in this member.
 *
 * Return: > -1  = number of entries in command map.
 *         -1    = error condition.
 *
 * Notes:  While WIL is expecting the number of entries in your 
 *         command table this is a good place to add entries to the 
 *         table if you are creating your table at runtime instead of 
 *         at DLL load time like this example does.
 *        
 ********************************************************************/
LONG CMath64::Initialization(HWND hWnd,           // hWnd of parent application.
                             HINSTANCE hInst,     // hInstance of parent application.
									  LPLONG lpLong,       // Pointer to your 32 bits.
									  LPVIPERVAR lpvvArg,  // Pointer to byte buffer.
									  LONG nArgCount)      // Size of lpvvArg in bytes.
{
	// You can also create extender table entries here.
	// Paramters are a pointer to a member, WIL function name, character length of name, 
	// required parameters, optional parameters, 0-16 paramter types.
	// For example,
	//m_ExtTable.push_back(EntryT<CMath64>(&CMath64::Add,_T("m64add"), 6, 2, 0, PTUni, PTUni)); 

	//return m_ExtTable.size();  // Number of extender implemented functions and constants.
	return _countof(m_ExtTable); // Number of extender implemented functions and constants.
}

/********************************************************************
 * Name: LoadTable  (case -2 : WILExtender2(W))
 *
 * Purpose: Case -2 is called repeatedly by the WIL interpreter.  On 
 *          each call, one CommandTable entry is passed back to the 
 *          WIL interpreter.  You can use this member implementation 
 *          in your own extender class by changing the type name in 
 *          used with the EntryT<> template class referenced in this
 *          member.
 *
 *
 * Return: 0 = done.
 *         1 = more entries.
 *
 * Notes:  
 *        
 ********************************************************************/
LONG CMath64::LoadTable(HWND hWnd,           // hWnd of parent application.
                        HINSTANCE hInst,     // hInstance of parent application.
								LPLONG lpLong,       // Pointer to your 32 bits.
								LPVIPERVAR lpvvArg,  // Pointer to table entry structure.
								LONG nArgCount)      // Index of entry to process.
{    
   
   if ( nArgCount >= static_cast <LONG> (_countof(m_ExtTable)) )
		return 0; // Done.
	
	//if ( nArgCount >= static_cast <LONG> (m_ExtTable.size()) )
   //   return 0; // Done.
   
	LPTTABLESTRUCT lpTable = reinterpret_cast<LPTTABLESTRUCT>(lpvvArg);  // Convenient source code simplification.
   
	if ( lpTable ) {

		const EntryT<CMath64> * pEntry = &(m_ExtTable[nArgCount]);

		// Make sure the name length is correct.
		assert(m_ExtTable[nArgCount].m_nSize == lstrlen(m_ExtTable[nArgCount].m_szName));

      lpTable->ident  = nArgCount;                    // Delimindex - must be consecutive!
		lpTable->type   = pEntry->m_nConstFunc;         // 2=FUNCTION 3=CONSTANT 4=COMMAND
      lpTable->param1 = pEntry->m_ArgCount.ParmCount; // (funct ? hi(reserved) lo(num-args))   (constant ? value)
      lpTable->ulMask = pEntry->m_ArgTypes.AllTypes;  // 64-bit mask (16 params, 4 bits each)
      lpTable->size   = pEntry->m_nSize;              // Length of function/constant name.

      // funct/const name - must be lowercase!
      ::lstrcpy(static_cast<LPTSTR>(lpTable->delim), pEntry->m_szName);
		::CharLower(static_cast<LPTSTR>(lpTable->delim));
   }
   
   return 1; // More entries.
}

/********************************************************************
 * Name: ShutDown  (case -3 : WILExtender2(W))
 *
 * Purpose: Case -3 is called during the WIL Shutdown process.  If 
 *          you need to do cleanup processing, add it to this member
 *          in your extender class.
 *
 * Return: 1 = success.
 *
 * Notes:  
 *        
 ********************************************************************/
LONG CMath64::ShutDown(HWND hWnd,           // hWnd of parent application.
                       HINSTANCE hInst,     // hInstance of parent application.
                       LPLONG lpLong,       // Pointer to your 32 bits.
                       LPVIPERVAR lpvvArg,  // Pointer to parameter array.
                       LONG nArgCount)      // Parameter count for lpvvArg.
{    
   
   return 1;  // OK.
}


