//////////////////////////////////////////////////////////////////////
// MyExtender.cpp: implementation of the CMyExtender class.
//
// This example demonstrates the use of the CWilExtender base 
// class to create an a new C++ based WIL extender. An extender 
// can be easily created if a few basic steps are followed:
//
// 1.  Define your extender class interface as derived from
//     the CWilExtender base class.  For example, 
//     class CMyExtender: public CWilExtender {... 
//     (see MyExteder.h)
//
// 2.  Include the macro  DECLARE_COMMAND_MAP(); somewhere 
//     within your class interface definition.
//     (see MyExteder.h)
//                                                                                                                                                                                                                      
// 3.  Make sure you create an instance of your class. For 
//     example, CMyExtender MyExtender;.
//
// 4.  Create a command map by using the BEGIN_COMMAND_MAP,
//     COMMAND_MAP_FUNCTION, COMMAND_MAP_CONSTANT, and
//     END_COMMAND_MAP macros. When using these macros
//     keep these points in mind:
//     - Include your class's name and the scope resolution 
//       operator (::) with a member name in the last 
//       parameter of the COMMAND_MAP_FUNCTION macro i.e.,
//       CMyExtender::MyMember.  
//     - Make sure NOT to place semi colons (;) after any macros. 
//     View the command map below for an example.

//     (These macros should look vaguely familiar to MFC 
//     programmers.)
//
// 5.  Implement class members to support each WIL extender
//     function. 
//     Note: All members referenced in the command map must 
//           accept the same five parameters and return BOOL.
// 
// Most of the CWilExtender members are declared virtual so those
// wishing to do advanced c++ programming can change class behavior
// without modifing the CWilExtender class directly.  If you do 
// choose to implement your own overrides of base class members, 
// in most cases, it is best to call the base class implemention 
// from within the override to insure that the WIL Interpreter 
// gets what it is expecting.
//////////////////////////////////////////////////////////////////////

#include <windows.h>      // Windows defines and prototypes.
#include <tchar.h>        // tchar defines and prototypes.
#include <objbase.h>      // COM defines and prototypes.
#define _CONVERSION_USES_THREAD_LOCALE
#include "atlbase.h"      // ATL suppport. (efficient string conversions...)
#include "resource.h"     // Defines for string ids.
#include "MyExtender.h"   // "this" class definition.
#include "assert.h"       // Document your assumptions.


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMyExtender::CMyExtender() 
{
   // This is a good place to set the extender's name...
   _tcsncpy( m_szExtName, _T("MyExtender"), sizeof(m_szExtName)-1);

   // ... and version.
   m_dwExtVersion = 20002;

   ::OleInitialize(NULL); // Garentees that COM is initialized.
}

// Create an instance of your extender object here.
CMyExtender MyExtender;

// Create the command table with one  entry per function or constant.
BEGIN_COMMAND_MAP(CMyExtender, MyExtender) 
//                                           Parm types          Num Parms  Func. Name    Implementation

COMMAND_MAP_FUNCTION( PARAM_TYPE(1,INTTYPE)|PARAM_TYPE(2,INTTYPE),    2,     "addint",         CMyExtender::AddInt)
COMMAND_MAP_FUNCTION( PARAM_TYPE(1,FLOATTYPE)|PARAM_TYPE(2,FLOATTYPE),2,     "addfloat",       CMyExtender::AddFloat)
COMMAND_MAP_FUNCTION( PARAM_TYPE(1,INTTYPE)|PARAM_TYPE(2,INTTYPE),    2,     "addlong",        CMyExtender::AddLong)
COMMAND_MAP_FUNCTION( PARAM_TYPE(1,STRTYPE)|PARAM_TYPE(2,STRTYPE),    2,     "catstr",         CMyExtender::CatStr)
COMMAND_MAP_FUNCTION( PARAM_TYPE(1,VTTYPE)|PARAM_TYPE(2,STRTYPE),     2,     "ismember",       CMyExtender::IsMember)
COMMAND_MAP_FUNCTION( PARAM_TYPE(1,STRTYPE),                          1,     "getvariable",    CMyExtender::GetVariable)
COMMAND_MAP_FUNCTION( PARAM_TYPE(1,STRTYPE)|PARAM_TYPE(2,INTTYPE),    2,     "makesavariable", CMyExtender::MakeSaVariable)
COMMAND_MAP_FUNCTION( PARAM_TYPE(1,STRTYPE)|PARAM_TYPE(2,INTTYPE),    2,     "getarrayelem",   CMyExtender::GetArrayElem)
COMMAND_MAP_FUNCTION( PT(1,STRTYPE)|PT(2,INTTYPE)|PT(3,STRTYPE),      3,     "setarraybstrelem",CMyExtender::SetArrayBstrElem)
COMMAND_MAP_FUNCTION( PARAM_TYPE(1,ASISTYPE),                         1,     "createvtdecimal",CreateVtDecimal)
COMMAND_MAP_FUNCTION( 0, 0, "testerror", CMyExtender::CauseError)
COMMAND_MAP_CONSTANT( 1, "@TestConstant")
END_COMMAND_MAP() 

CMyExtender::~CMyExtender()
{

   // Perform any cleanup here.
   ::OleUninitialize();
}

//////////////////////////////////////////////////////////////////////
// Members that implement extender functions.
//////////////////////////////////////////////////////////////////////

////////////////////
// Add two integers.
////////////////////
BOOL CMyExtender::AddInt(HWND hWnd, 
                         HINSTANCE hInst, 
                         LPLONG lpLong, 
                         LPVIPERVAR lpvvArg, 
                         LONG nArgCount)
{
    assert(nArgCount == 2);

    int nSum = lpvvArg[0].x +  lpvvArg[1].x;

    return ReturnResult(nSum);
}

////////////////////
// Add two floating point numbers.
////////////////////
BOOL CMyExtender::AddFloat(HWND hWnd, 
                           HINSTANCE hInst, 
                           LPLONG lpLong, 
                           LPVIPERVAR lpvvArg, 
                           LONG nArgCount)
{
   assert(nArgCount == 2);
    
   double dSum = lpvvArg[0].flt +  lpvvArg[1].flt;

   return ReturnResult(dSum);
}

////////////////////
// Add two long integers numbers.
////////////////////
BOOL CMyExtender::AddLong(HWND hWnd, 
                          HINSTANCE hInst, 
                          LPLONG lpLong, 
                          LPVIPERVAR lpvvArg, 
                          LONG nArgCount)
{
   assert(nArgCount == 2);
   
   LONG lSum = lpvvArg[0].x +  lpvvArg[1].x;
   
   return ReturnResult(lSum);
}

////////////////////
// Concatenate two strings.
////////////////////
BOOL CMyExtender::CatStr(HWND hWnd, 
                         HINSTANCE hInst, 
                         LPLONG lpLong, 
                         LPVIPERVAR lpvvArg, 
                         LONG nArgCount)
{
   assert(nArgCount == 2);

   int nSize = strlen(lpvvArg[0].lpstr) + strlen(lpvvArg[1].lpstr) + 1;
   
   // Create a string buffer.
   LPTSTR lpBigger = new char [nSize];
   
   // A little error checking.
   // Note: "RESULT_MINOR_ERROR" is from addons.h.
   if ( lpBigger== NULL )
      return ReturnError(RESULT_MINOR_ERROR, IDS_NOMEMORY);
   
   // Concatenate the input strings together.
   strcpy(lpBigger,lpvvArg[0].lpstr);
   strcat(lpBigger,lpvvArg[1].lpstr);

   // Send the results to WIL but do not return immediately.
   int nReturn = ReturnResult(lpBigger);
   
   // Clean up
   delete [] lpBigger;
   
   // Now return with value supplied by the WIL Interpreter.
   return nReturn;
}


/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
//  The following members are generally not very useful but 
//  do serve to illustrate the features added with extender 
//  level 1004.
//  (They come form an extender used as part of the WinBatch 
//   regression test suite.)
////////////////////////////////////////////////////////////

////////////////////
// Determine if string is name of member of COM object.
// Script parameters: 1. Variant of type VT_DISPATCH - COM Object
//                    2. String - Name of method or property.
// 
////////////////////
BOOL CMyExtender::IsMember(HWND hWnd, 
                           HINSTANCE hInst, 
                           LPLONG lpLong, 
                           LPVIPERVAR lpvvArg, 
                           LONG nArgCount)
{
   assert(nArgCount == 2);

   if ( V_VT(&(lpvvArg[0].variant)) != VT_DISPATCH )
      return ReturnError(RESULT_MINOR_ERROR, IDS_NOTANOBJECT);
   
   // Get an IDispatch interface pointer.
   LPDISPATCH pDisp = V_DISPATCH(&(lpvvArg[0].variant));
   if ( pDisp == NULL )
      return ReturnError(RESULT_MINOR_ERROR, IDS_NOTANOBJECT);

   // Convert the second parameter to a OLESTR
#if _MSC_VER < 1300  // Not Visual Studio .NET - ATL
   USES_CONVERSION;  // Visual Studio 6 - ATL
   LPOLESTR pNames = T2OLE(lpvvArg[1].lpstr);  
#else
   USES_CONVERSION_EX;  // Visual Studio .NET - ATL
   LPOLESTR pNames = T2OLE_EX(lpvvArg[1].lpstr, _ATL_SAFE_ALLOCA_DEF_THRESHOLD);  
#endif
   
   // Look up the potential member name.
   LONG          nMemId;    // Out value not used. 
   HRESULT hr = pDisp->GetIDsOfNames(IID_NULL, &pNames, 1, LOCALE_USER_DEFAULT, &nMemId);

   // Return TRUE or FALSE using handy macro.
   return ReturnResult(SUCCEEDED(hr));
}

////////////////////
// Get the value of a named variable.
// Script parameters: 1. String - Name of WIL variable
//
////////////////////
BOOL CMyExtender::GetVariable(HWND hWnd, 
                              HINSTANCE hInst, 
                              LPLONG lpLong, 
                              LPVIPERVAR lpvvArg, 
                              LONG nArgCount)
{
   assert(nArgCount == 1);
   vipervar vV;  // Holds variables contents.

   // Query WB for the value of the input variable.
   if ( DllVarHandler(lpvvArg[0].lpstr, &vV,  1) ) 
      return ReturnError(RESULT_MINOR_ERROR, IDS_VARVALUEFAILED);

   // Send the results to WIL but do not return immediately.
   int nReturn = ReturnResult(&vV); 

   // Variants are always a copy so we are responsible for cleanup.
   if ( vV.inittype == VARTYPE_VARIANT )
     VariantClear(&(vV.variant));

   // Now return with value supplied by the WIL Interpreter.
   return nReturn;
}

////////////////////
// Create a one dimension safearray of variants variable.
// Script parameters: 1. String - Name for new WIL variable;
//                    2. Number - Number elements in new safearray.
//
////////////////////
BOOL CMyExtender::MakeSaVariable(HWND hWnd, 
                                 HINSTANCE hInst, 
                                 LPLONG lpLong, 
                                 LPVIPERVAR lpvvArg, 
                                 LONG nArgCount)
{
   assert(nArgCount == 2);
   vipervar Vv;  // Holds New variable's contenst.
   
   // Make a one dimentional array.
   SAFEARRAYBOUND rgsaBounds[1];  
   rgsaBounds[0].lLbound = 0;
   rgsaBounds[0].cElements = lpvvArg[1].x;
   
   SAFEARRAY * psaArray = SafeArrayCreate(VT_VARIANT, 1, rgsaBounds);
   if ( !psaArray ) 
      return ReturnError(RESULT_MINOR_ERROR, IDS_ARRAYCREATEFAIL);

   // Initialize array.
   VARIANT Variant;
   VariantInit(&Variant);
   HRESULT hr = S_OK;
   for ( long i = 0; i<lpvvArg[1].x; i++ ) {    
      hr = SafeArrayPutElement(psaArray, &i, (LPVOID)(&Variant));
      if ( FAILED(hr) )
         break;
   }

   if ( FAILED(hr) ) {

      SafeArrayDestroy(psaArray);
      return ReturnError(RESULT_MINOR_ERROR, IDS_ARRAYINITFAIL);
   }
   
   // Create a safe array viper variable.
   Vv.inittype          = VARTYPE_VARIANT;
   V_VT(&Vv.variant)    = VT_ARRAY|VT_VARIANT;
   V_ARRAY(&Vv.variant) = psaArray;

   // Set the new safearray as a variable with the specified name.
   if ( DllVarHandler(lpvvArg[0].lpstr, &Vv, 2) ) {
      SafeArrayDestroy(psaArray);
      return ReturnError(RESULT_MINOR_ERROR, IDS_VARNOTFOUND);
   }

   // WB owns a copy of the array now so can delete ours.
   SafeArrayDestroy(psaArray);

   // Return success.
   return ReturnResult(TRUE); 
   
 }

////////////////////
// Get the value of an WIL array element.
// Script parameters: 1. String - Name of will array variable.
//                    2. Number - Index of array element to return.
//
////////////////////
BOOL CMyExtender::GetArrayElem(HWND hWnd, 
                               HINSTANCE hInst, 
                               LPLONG lpLong, 
                               LPVIPERVAR lpvvArg, 
                               LONG nArgCount)
{
   assert(nArgCount == 2);
   vipervar vArray;  // Holds array variables contents.
   vipervar vElem;   // Array elements value
   
   if ( DllVarHandler(lpvvArg[0].lpstr, &vArray,  1) ) 
      return ReturnError(RESULT_MINOR_ERROR, IDS_VARVALUEFAILED);

   if ( vArray.inittype != VARTYPE_ARRAY ) 
      return ReturnError(RESULT_MINOR_ERROR, IDS_VARNOTARRAY);

   // Query WB for the value of the requested element.
   if ( DllArrHandler( &vArray, lpvvArg[1].x, 0, 0, 0, 0, &vElem, 1) )
      return ReturnError(RESULT_MINOR_ERROR, IDS_VARVALUEFAILED);

   // Send the results to WIL but do not return immediately.
   int nReturn = ReturnResult(&vElem); 

   // Clean up variants unless they come from lpvvArg.
   if ( vElem.inittype == VARTYPE_VARIANT )
      VariantClear(&(vElem.variant));

   return nReturn;
}

////////////////////
// Set the value of a WIL array element to VT_BSTR value.
// Script parameters: 1. String - Name of will array variable.
//                    2. Number - Index of array element to return.
//                    3. String - text to convert to a BSTR and assign to array element.
////////////////////
BOOL CMyExtender::SetArrayBstrElem(HWND hWnd, 
                                   HINSTANCE hInst, 
                                   LPLONG lpLong, 
                                   LPVIPERVAR lpvvArg, 
                                   LONG nArgCount)
{
   assert(nArgCount == 3);
   vipervar vArray;  // Holds array variable's contents.
   vipervar vElem;   // Array element's value.
   
   if ( DllVarHandler(lpvvArg[0].lpstr, &vArray,  1) ) 
      return ReturnError(RESULT_MINOR_ERROR, IDS_VARVALUEFAILED);

   if ( vArray.inittype != VARTYPE_ARRAY ) 
      return ReturnError(RESULT_MINOR_ERROR, IDS_VARNOTARRAY);

   // Create a variant bstr WIL variable.
   int nConvertedLen = ::MultiByteToWideChar(CP_ACP, 0, lpvvArg[2].lpstr,-1, NULL, NULL);
	BSTR bs = ::SysAllocStringLen(NULL, nConvertedLen);

   if ( bs != NULL ){

      // Needs error checking.
      ::MultiByteToWideChar(CP_ACP, 0, lpvvArg[2].lpstr, -1, bs, nConvertedLen);
   }

   vElem.inittype           = VARTYPE_VARIANT;
   V_VT(&(vElem.variant))   = VT_BSTR;
   V_BSTR(&(vElem.variant)) = bs;
   
   // Set the requested array value.
   if ( DllArrHandler( &vArray, lpvvArg[1].x, 0, 0, 0, 0, &vElem, 2) ) {
      ::VariantClear(&(vElem.variant));
      return ReturnError(RESULT_MINOR_ERROR, IDS_VARVALUEFAILED);
   }

   // Always clean up variants unless they come from lpvvArg.
   ::VariantClear(&(vElem.variant));

   return ReturnResult(TRUE); 
}

////////////////////
// Creates a Variant of type VT_DECIMAL from input.
// Script parameters: 1. String/integer/float/variant - value to turn into a VT_DECIMAL
//
// Would be easier to let the interpreter convert to a string for us but...
////////////////////
BOOL CMyExtender::CreateVtDecimal(HWND hWnd, 
                                 HINSTANCE hInst, 
                                 LPLONG lpLong, 
                                 LPVIPERVAR lpvvArg, 
                                 LONG nArgCount)
{
   assert(nArgCount == 1);
   VARIANT  vDecimal;

   VariantInit(&vDecimal);

   // Set variant based on WIL variable type.
   if ( lpvvArg[0].inittype == VARTYPE_STRING ) {

      // Convert to bs.
      int nConvertedLen = ::MultiByteToWideChar(CP_ACP, 0, lpvvArg[0].lpstr,-1, NULL, NULL);
	   BSTR bs = ::SysAllocStringLen(NULL, nConvertedLen);

      if ( bs != NULL ){

         // Needs error checking.
         ::MultiByteToWideChar(CP_ACP, 0, lpvvArg[0].lpstr, -1, bs, nConvertedLen);
      }
      
      V_VT(&vDecimal)   = VT_BSTR;
      V_BSTR(&vDecimal) = bs;
   }
   else if ( lpvvArg[0].inittype == VARTYPE_FLOATNUM ) {

      V_VT(&vDecimal) = VT_R8;
      V_R8(&vDecimal) = lpvvArg[0].flt;
   }
   else if ( lpvvArg[0].inittype == VARTYPE_INT ) {
      
      V_VT(&vDecimal) = VT_I4;
      V_I4(&vDecimal) = lpvvArg[0].x;
   }
   else if ( lpvvArg[0].inittype == VARTYPE_VARIANT ) {
      
      vDecimal = lpvvArg[0].variant;
   }
   else {
   
      return ReturnError(RESULT_MINOR_ERROR, IDS_CONVERTFAILED);
   }

   // Perform the conversion to VT_DECIMAL.
   BOOL bResult = FALSE;
   HRESULT hr = ::VariantChangeType(&vDecimal, &vDecimal, 0, VT_DECIMAL);
   if ( SUCCEEDED(hr) ) 
      bResult = ReturnResult(&vDecimal); 
   
   // Always clean up variants unless they come from lpvvArg.
   ::VariantClear(&(vDecimal));

   if ( FAILED(hr) ) 
      ReturnError(RESULT_MINOR_ERROR, IDS_CONVERTFAILED);

   return bResult;
}

////////////////////
// Generate an error message.
////////////////////
BOOL CMyExtender::CauseError(HWND hWnd, 
                             HINSTANCE hInst, 
                             LPLONG lpLong, 
                             LPVIPERVAR lpvvArg, 
                             LONG nArgCount)
{
   // Note: "RESULT_MINOR_ERROR" is from addons.h.
   return ReturnError(RESULT_MINOR_ERROR, IDS_ERROR_TEST);
}

