// WilExtender.cpp: implementation of the CWilExtender class.
//
//////////////////////////////////////////////////////////////////////
#include <windows.h>      // Windows defines and prototypes.
#include <tchar.h>        // tchar defines and prototypes.
#include "WilExtender.h"  // Definition of this class.


//////////////////////////////////////////////////////////////////////
// Static class properties.
//////////////////////////////////////////////////////////////////////
HINSTANCE         CWilExtender::m_hInstDLL;  // DLL instance handle.

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CWilExtender::CWilExtender()
{
   m_lpWork                            = NULL;             
   m_dwWorkSize                        = 0; 
   m_dwExtVersion                      = EXTENDERVERSION;
   m_szExtName[sizeof(m_szExtName)-1]  = 0;

   // Set a default extender name.
   _tcsncpy( m_szExtName, _T("WilExtender"), sizeof(m_szExtName)-1);
   
}

CWilExtender::~CWilExtender()
{
   // Clean up work buffer.
   if ( m_lpWork )
      delete [] m_lpWork;
}


///////////////////////////////////////////////////////////////////////
//  Supporting members use by WIL to startup and shutdown the extender.
// (May override in your derived class.)
///////////////////////////////////////////////////////////////////////

/********************************************************************
 * Name: Initalization  (case -1)
 *
 * Purpose: Case -1 is called once, when your extender is loaded.  
 *          Override this member to perform your one-time initialization 
 *          tasks.  The return value should be the number of items 
 *          in your command map (functions and constants).  You may
 *          also return 0 if you are unable to determine the correct 
 *          number of items. You may return a negative value (-1) to 
 *          indicate that your extender was unable to successfully 
 *          initialize, for example, if it is running on an unsupported 
 *          platform or a dependent DLL could not be loaded.  This will
 *          cause the AddExtender command to fail.  You may additionally 
 *          provide a  string indicating the reason for the failure:
 *          lpvvArg points to a character buffer into which you may 
 *          copy an error message.
 *          "nArgCount" is the size of the buffer pointed to by lpvvArg, 
 *          in bytes. You can expect this to be a 256-byte buffer.
 *
 * Return: > -1  = number of entries in command map.
 *         -1    = error condition.
 *
 * Notes:  Since WIL is expecting the number of entries in your 
 *         command table map as a return value, it is recommended 
 *         that you call this member and return its result in
 *         your override member.
 *        
 ********************************************************************/
BOOL CWilExtender::Initalization(HWND hWnd,           // hWnd of parent application.
                                 HINSTANCE hInst,     // hInstance of parent application.
                                 LPLONG lpLong,       // Pointer to your 32 bits.
                                 LPVIPERVAR lpvvArg,  // Pointer to byte buffer.
                                 LONG nArgCount)      // Size of lpvvArg in bytes.
{
   BOOL bInterfaceOk = FALSE;
   
   if ( m_lpViperStruct->dwLevel >= EXTENDER_FEATURE_LEVEL ) {
      
      bInterfaceOk = TRUE;
   }
   
   if ( bInterfaceOk == FALSE )  {
      
      lstrcpy((LPTSTR)lpvvArg,_T("Need newer version of WIL DLL"));
      return -1;  // Error.
   }
   
   // Get the function map.
   LPCOMMAND_MAP pMap = GetCommandMap();
   if ( !pMap )
      return -1;  // Error.

   // Count then number of entries in the command table map.
   for ( int nCount = 0; pMap[nCount].t.ident != -1; nCount++);

   return (nCount);  // Number of entries.
}


/********************************************************************
 * Name: LoadTable  (case -2)
 *
 * Purpose: Case -2 is called repeatedly by the WIL extender.  On 
 *          each call, one CommandTable entry is passed back to the 
 *          WIL extender.  It is unlikely that you will need to override 
 *          this member in your derived extender class.
 *
 *
 * Return: 0 = done.
 *         1 = more entries.
 *
 * Notes:  
 *        
 ********************************************************************/
BOOL CWilExtender::LoadTable(HWND hWnd,           // hWnd of parent application.
                             HINSTANCE hInst,     // hInstance of parent application.
                             LPLONG lpLong,       // Pointer to your 32 bits.
                             LPVIPERVAR lpvvArg,  // Pointer to table entry structure.
                             LONG nArgCount)      // Index of entry to process.
{         
   LPTABLESTRUCT lpTable;
   UINT          j;
   
    // Get a pointer to this function map.
   LPCOMMAND_MAP pMap = GetCommandMap();
   if ( !pMap )
      return 0;   // No map to process.

   lpTable = (LPTABLESTRUCT) lpvvArg;
   
   j = (UINT) nArgCount;
   
   if (pMap[j].t.ident == -1)
      return (0); // Done.
   
   if (lpTable) {

      lpTable->ident  = pMap[j].t.ident = nArgCount; // delimindex - must be consecutive!
      lpTable->type   = pMap[j].t.type;              // 2=FUNCTION 3=CONSTANT 4=COMMAND
      lpTable->param1 = pMap[j].t.param1;            // (funct ? hi(reserved) lo(num-args))   (constant ? value)
      lpTable->ulMask = pMap[j].t.ulMask;            // 64-bit mask (16 params, 4 bits each)
      lpTable->size   = lstrlen((LPSTR) pMap[j].t.delim); // length of function name
      
      ::CharLower(pMap[j].t.delim);

      // funct/const name - must be lowercase!
      ::lstrcpy((LPSTR) lpTable->delim, (LPSTR)  pMap[j].t.delim);
   }
   
   return (1); // More entries.
}

/********************************************************************
 * Name: ShutDown  (case -3)
 *
 * Purpose: Case -3 is called during the WIL Shutdown process.  If 
 *          you need to do cleanup processing, override this member
 *          in your derived extender class.
 *
 * Return: 1 = success.
 *
 * Notes:  
 *        
 ********************************************************************/
BOOL CWilExtender::ShutDown(HWND hWnd,           // hWnd of parent application.
                            HINSTANCE hInst,     // hInstance of parent application.
                            LPLONG lpLong,       // Pointer to your 32 bits.
                            LPVIPERVAR lpvvArg,  // Pointer to parameter array.
                            LONG nArgCount)      // Parameter count for lpvvArg.
{    
   
   return (1);  // OK.
}



///////////////////////////////////////////////////////////////////////////
//  Members that return results and error messages to the WIL Interpreter.
///////////////////////////////////////////////////////////////////////////

/********************************************************************
 * Name: ReturnResult - long integer
 *
 * Purpose: Passes long integer result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
BOOL CWilExtender::ReturnResult( LONG lNumber ) // In - Long integer.
{
   m_lpViperReturn->x        = lNumber;
   m_lpViperReturn->inittype = VARTYPE_INT;

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnResult - integer
 *
 * Purpose: Passes integer result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  Inlined in CWilExtender.h
 *        
 ********************************************************************/
//BOOL CWilExtender::ReturnResult( int nNumber ) // In - integer.

/********************************************************************
 * Name: ReturnResult - double
 *
 * Purpose: Passes double result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
BOOL CWilExtender::ReturnResult(double dNumber) // In - floating point number.
{
   m_lpViperReturn->flt      = dNumber;
   m_lpViperReturn->inittype = VARTYPE_FLOATNUM;

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnResult - string
 *
 * Purpose: Passes string result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
BOOL CWilExtender::ReturnResult(LPTSTR lpStr) // In - pointer to string
{
   m_lpViperReturn->lpstr      = lpStr;
   m_lpViperReturn->inittype = VARTYPE_STRING;

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnResult - variant
 *
 * Purpose: Passes variant result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
BOOL CWilExtender::ReturnResult(LPVARIANT lpVariant) // In - pointer to variant
{
   m_lpViperReturn->inittype = VARTYPE_VARIANT;
   m_lpViperReturn->variant  = *lpVariant;   // Note: shallow copy.

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnResult - vipervar
 *
 * Purpose: Passes vipervar result to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes:  
 *        
 ********************************************************************/
BOOL CWilExtender::ReturnResult(LPVIPERVAR lpVv) // In - pointer to vipervar
{
   m_lpViperReturn->flt      = lpVv->flt;
   m_lpViperReturn->handle   = lpVv->handle;
   m_lpViperReturn->inittype = lpVv->inittype;
   m_lpViperReturn->lpstr    = lpVv->lpstr;
   m_lpViperReturn->variant  = lpVv->variant;

   return m_lpViperCallback(RESULT_SUCCESS, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnError 
 *
 * Purpose: Passes severity code and error message to the WIL Interpreter.
 *
 * Return: Integer - use as return value of your member function.   
 *
 * Notes: nSeverity -  MUST be -1, -2 or -3.
 *                     Use -1 for most run of the mill errors.
 *                         -2 for serious errors
 *                         -3 for absolutely fatal "we better die" errors
 *
 *        nErrorID - is the string ID from the RC file that contains 
 *                   the error message.  The first three characters of
 *                   the string should be the same as the string ID, 
 *                   and the fourth character should be a colon.
 *        
 ********************************************************************/
BOOL CWilExtender::ReturnError( int nSeverity, // In- severity code.
                                int nErrorID)  // In- resource id.
{
   // Load the error message.
   m_lpViperReturn->lpstr    = (char *) GetWorkBuffer(MIN_MESSAGE_BUFFER);
   m_lpViperReturn->inittype = VARTYPE_STRING;
   LoadMyExtString( nErrorID, m_lpViperReturn->lpstr, GetWorkSize());

   return m_lpViperCallback(nSeverity, 
                            m_ViperInterfaceVersion,                     
                            m_lpViperStruct,                             
                            m_lpViperReturn);
}

/********************************************************************
 * Name: ReturnCancel
 *
 * Purpose: Supplies value that WIl Interpreter understands to
 *          to mean that the current operation is to be cancelled.
 *          This value can be used as a return value of a class
 *          member. For example,
 *          BOOL MySubClass::MyMember(....) { 
 *            .....
 *            if ( bNeverMind == TRUE )
 *             return ReturnCancel();
 *          }
 *
 * Return: CANCEL 
 *
 * Notes:  Inlined in CWilExtender.h
 *        
 ********************************************************************/
//virtual BOOL ReturnCancel()

///////////////////////////////////////////////////////////////////////////
//  Utility members
///////////////////////////////////////////////////////////////////////////

/********************************************************************
 * Name: GetWorkBuffer
 *
 * Purpose: This member returns a pointer to the class work buffer.
 *          The work buffer is adjusted to be at least as big as 
 *          the requested size.
 *          The buffer is zeroed unless the bSaveConent parameter is 
 *          TRUE.  
 *
 * Return: Pointer to VOID on success or NULL on error.
 *
 * Notes:   
 *        
 ********************************************************************/
VOID * CWilExtender::GetWorkBuffer(DWORD dwSize /*= 0x100*/,        //In-required memory
                                   BOOL bSaveContent /* = FALSE */) //In-TRUE = preserve content. 
{
 
   // Increase the size?
   if ( m_dwWorkSize < dwSize ) {
      VOID * lpTemp = (VOID *) new unsigned char[dwSize];

      if ( lpTemp ) {

         // Copy memory if necessary.
         if ( bSaveContent && m_dwWorkSize ) 
            ::CopyMemory(lpTemp, m_lpWork, m_dwWorkSize);
      
         m_dwWorkSize = dwSize;
      }
      else
         m_dwWorkSize = 0;

      if ( m_lpWork ) 
         delete [] m_lpWork;
      
      m_lpWork = lpTemp;
   }

   return m_lpWork;
}

/********************************************************************
 * Name: LoadMyExtString
 *
 * Purpose: Loads string from resource or language file.  
 *
 * Return: Number of chars in output.
 *
 * Notes:  
 *        
 ********************************************************************/
int CWilExtender::LoadMyExtString(int wID, LPTSTR lpBuffer, int nBuffermax)
{
   LPTSTR       lpSlash;
   int          nCharsBack, i;
   static TCHAR szLangIni[MAX_PATH];
   static int  nIniExist       = 0;           // 0=uninit   1=notrans   2=xlate it
   static TCHAR szLangSection[MAX_EXT_NAME*2]; 
   static TCHAR szLANGDLL[]     = "WWWELANG";
   static TCHAR szXXX[]         = "XXX";
   static TCHAR szMAIN[]        = "MAIN";
   static TCHAR szINTL[]        = "INTL";
   static TCHAR szLibIni[]      = "WWWBATCH.INI";
   static TCHAR szPanic[256];
   TCHAR        SafeBuffer[20];
   TCHAR        szKey[20];
   
   OutputDebugString("In LoadMyExtString\n");  

   // If we errored allocating a buffer, we will use our panic buffer.
   if ( !nBuffermax ) {
      lpBuffer = szPanic;
      nBuffermax = sizeof(szPanic);
   }
   
   if (nIniExist == 0 || wID == 0)  {    // Process not initialized
   
      OFSTRUCT OFS;
      HANDLE   hRes;
      
      nIniExist    = 1;
      szLangIni[0] = '\0';
      
      // First see if a language override has been installed
      ::GetPrivateProfileString(szMAIN, "Language", szXXX, SafeBuffer, 
                                sizeof(SafeBuffer), szLibIni);
      
      if (lstrcmp(SafeBuffer, szXXX) == 0)  // no language override
         ::GetProfileString(szINTL, "sLanguage", "enu", SafeBuffer, sizeof(SafeBuffer));  // look foreign lang.  Def to enu
      
      // Get DLL location for search for our lng dll
      i = ::GetModuleFileName(CWilExtender::m_hInstDLL, szLangIni, sizeof(szLangIni));
      
      if (i == 0)
         szLangIni[0] = '\0';
      else {
         lpSlash = MyLastSlashInString(szLangIni);
         
         if (lpSlash)
            *(AnsiNext(lpSlash)) = '\0';
         else
            szLangIni[0] = '\0';
      }
      
      lstrcat(szLangIni, szLANGDLL);     // glue on WWWELANG
      lstrcat(szLangIni, ".");           // glue on .
      lstrcat(szLangIni, SafeBuffer);    // glue on enu or whatnot
      
      hRes = (HANDLE) ::OpenFile(szLangIni, &OFS, OF_EXIST);  // well, does it exist?
      
      if (hRes == (HANDLE) -1) {
         wsprintf((LPSTR) szLangIni, (LPSTR) "%s.%s", (LPSTR) szLANGDLL, (LPSTR) SafeBuffer);
         hRes = (HANDLE) ::OpenFile(szLangIni, &OFS, OF_EXIST);
         
         if (hRes != (HANDLE) -1)
            lstrcpyn(szLangIni, OFS.szPathName, sizeof(szLangIni));
      }
      
      if (hRes != (HANDLE) -1) {

         // Create the extenders section name from the extender name.
         _tcsncpy(szLangSection, m_szExtName, sizeof(szLangSection)-1);
         _tcsncat(szLangSection, _T(" Extender"), MAX_EXT_NAME);
         nIniExist = 2;
      }
   }
   
   if ( (wID >= 100) && (nIniExist == 2) ) {
      szKey[0] = 'X';
      szKey[1] = '\0';
      _ltoa(wID, szKey + 1, 10);
      

      nCharsBack = ::GetPrivateProfileString(szLangSection, szKey, szXXX, lpBuffer, nBuffermax, szLangIni);
      
      if (lstrcmp(lpBuffer, szXXX) == 0) {
         lpBuffer[0] = '\0';
         nCharsBack = ::LoadString(CWilExtender::m_hInstDLL, wID, lpBuffer, nBuffermax);
      }
   }
   else {
      lpBuffer[0] = '\0';
      nCharsBack = ::LoadString(CWilExtender::m_hInstDLL, wID, lpBuffer, nBuffermax);
   }
      
   while (*lpBuffer) {
      if (*lpBuffer == '~')
         *lpBuffer = '\n';
      
      if (*lpBuffer == '|')
         *lpBuffer = '\t';
      
      lpBuffer = ::AnsiNext(lpBuffer);
   }
   
   return nCharsBack;
}

/********************************************************************
 * Name: MyLastSlashInString
 *
 * Purpose: Returns a pointer to the last slash (forward or 
 *          backward) found in a string. Supports DBCS.
 *
 * Return: Pointer to the last slash.
 *
 * Notes:  
 *        
 ********************************************************************/
LPTSTR CWilExtender::MyLastSlashInString(LPTSTR lpString)
{
  LPTSTR lpPtr, lpSlash;

  lpSlash = NULL;

  for (lpPtr = lpString; *lpPtr; lpPtr = ::CharNext(lpPtr))
    if ((*lpPtr == _T('\\')) || (*lpPtr == _T('/')))
      lpSlash = lpPtr;

  return (lpSlash);
}

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
//  DLL Exports required by WIL. (You most likely will not need to change.)
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

/* =========================================================================
                                    DllMain
   ========================================================================= */

extern "C" BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{
   
   UNREFERENCED_PARAMETER(dwReason);
   UNREFERENCED_PARAMETER(lpReserved);
   
   // Save the instance handle.
   CWilExtender::m_hInstDLL = (HINSTANCE)hInst;

   return (TRUE);
}

/*===========================================================================

  WILExtenderQuery
  
  WILExtenderQuery is an optional (but recommended) entry point.  It is used
  to provide helpful information to the WIL DLL and possibly other callers.
    
===========================================================================*/

extern "C" DllExport LONG FAR PASCAL WILExtenderQuery(short  nRequest, // type of information being requested
                                                      LPVOID lpvInfo,  // information buffer
                                                      LONG   nCount)   // additional data
{
   
   LONG        i;
   
   // Make sure we have an object.
   if ( !CWilExtender::m_pThisObject ) 
      return 0;  // Could not handle request.

   switch ( nRequest ) {
      

   /*---------------------------------------------------------------------------
   
     case 1 is used to identify yourself to a caller.
     If lpvInfo is not NULL, it points to a character buffer where you may copy
     a string describing the extender.
     nCount is the size of the buffer pointed to by lpvInfo, in bytes.  When
     called by the WIL DLL, you can expect this to be a 256-byte buffer.
     The return value should be the version number of the extender.
     
      ---------------------------------------------------------------------------*/
   case 1:                            // version information
      {
         
         if ((lpvInfo) && (nCount > 0)) {
            
            TCHAR *lpWork = (TCHAR *) CWilExtender::m_pThisObject->GetWorkBuffer();
            if ( lpWork ) {
               CWilExtender::m_pThisObject->LoadMyExtString(1000, 
                                                            lpWork, 
                                                            CWilExtender::m_pThisObject->GetWorkSize());
               lstrcpyn((LPSTR) lpvInfo, lpWork, nCount);
            }
         }

         // Use the class's version member.
         return CWilExtender::m_pThisObject->m_dwExtVersion;
      }
      
      
   /*---------------------------------------------------------------------------
      
        case 2 is used to inform the WIL DLL of your preferred interface level.
        lpvInfo points to an array of LONG's, indicating the interface levels
        supported by the WIL DLL (each element represents one level).
        nCount indicates the number of elements in the array pointed to by lpvInfo.
        
        The return value must be the same as one of the elements in the lpvInfo
        array.  Currently, the WIL DLL supports interface levels 1 and 2, so your
        return value should be 1 or 2.  Newly designed extenders (such as WILX)
        use interface level 2 and have a WilExtender2 entry point.  However, it
        is possible to design an extender to support multiple interface levels.
          
        If you return an invalid value, the extender will not be loaded and the
        AddExtender command will fail.
            
     ---------------------------------------------------------------------------*/
   case 2:                            // preferred interface version
      
      {
         
         for (i = 0; i < nCount; i++)
            if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
               return (INTERFACELEVEL);
            
         return (-1);                     // didn't find any we like
      }
      
      
   /*---------------------------------------------------------------------------
      
        Return 0 for any unknown requests, or any requests you don't handle!
        
     ---------------------------------------------------------------------------*/
   default:
      
      return (0);
   }  
}

/* =========================================================================
   WIL Extender main entry point.

   All functionality is accessed from here.
   
   ========================================================================= */
                    
extern "C" DllExport LONG FAR PASCAL WILExtender2
                 (HWND            hWnd,                   // hWnd of parent application
                 HINSTANCE       hInst,                   // hInstance of parent application
                 LONG            ViperInterfaceVersion,   // WIL interface version
                 short           CmdId,                   // identifies extender function being called
                 LPVIPERCALLBACK lpViperCallback,         // pointer to return value callback function
                 LPVIPERSTRUCT   lpViperStruct,           // pointer to information structure
                 LPLONG          lpLong,                  // pointer to your 32 bits
                 LPVIPERVAR      lpvvArg,                 // pointer to parameter array
                 LONG            nArgCount,               // parameter count for lpvvArg
                 LPVIPERVAR      lpViperReturn)           // pointer to return value structure
{
 

   // Call the member that supports the extender request. 
   if ( CWilExtender::m_pThisObject ) {

      // Set member properties for use by derived classes and 
      // return methods.
      CWilExtender::m_pThisObject->m_ViperInterfaceVersion = ViperInterfaceVersion;  
      CWilExtender::m_pThisObject->m_lpViperCallback       = lpViperCallback;        
      CWilExtender::m_pThisObject->m_lpViperStruct         = lpViperStruct;          
      CWilExtender::m_pThisObject->m_lpViperReturn         = lpViperReturn; 
      CWilExtender::m_pThisObject->m_CmdId                 = CmdId;

      
      // Check for an initialization / termination request.
      switch (CmdId) {
      case -1: return CWilExtender::m_pThisObject->Initalization(hWnd, hInst, lpLong, lpvvArg, nArgCount);
      case -2: return CWilExtender::m_pThisObject->LoadTable(hWnd, hInst, lpLong, lpvvArg, nArgCount);
      case -3: return CWilExtender::m_pThisObject->ShutDown(hWnd, hInst, lpLong, lpvvArg, nArgCount);
      default: 
            
         // Extender functions.

         // Get the function/constant/command table mapping structure.
         LPCOMMAND_MAP pMap = CWilExtender::m_pThisObject->GetCommandMap();

         // Use the derived class's member to process request.
         return (CWilExtender::m_pThisObject->*(pMap[CmdId].pMember))(hWnd,hInst,lpLong,lpvvArg,nArgCount);
      }
   }
   else {

      return FALSE;  // No derived class so error.
   }
}

///////////////////////////////////////////////////////////////////////
//  End Exported Functions.
///////////////////////////////////////////////////////////////////////
